# (c) 2020-2022 MPIB <https://www.mpib-berlin.mpg.de/>,
#
# This file is part of castellum-scheduler.
#
# castellum-scheduler is free software; you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# Castellum is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with Castellum. If not, see
# <http://www.gnu.org/licenses/>.

from datetime import timedelta
from collections import OrderedDict

from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib import messages
from django.db import models
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import CreateView
from django.views.generic import DeleteView
from django.views.generic import ListView
from django.views.generic import UpdateView
from django.views.generic import View

from .forms import ScheduleForm
from .models import Invitation
from .models import Schedule


class OwnerRequiredMixin(LoginRequiredMixin):
    def get_queryset(self):
        return self.request.user.schedule_set.order_by('pk')


class ScheduleListView(OwnerRequiredMixin, ListView):
    model = Schedule
    paginate_by = 20


class ScheduleFormMixin:
    model = Schedule
    form_class = ScheduleForm

    def get_success_url(self):
        return reverse('schedule-update', args=[self.object.uuid])

    def form_valid(self, form, *args):
        messages.success(self.request, _('Data has been saved.'))
        return super().form_valid(form, *args)


class ScheduleCreateView(LoginRequiredMixin, ScheduleFormMixin, CreateView):
    def form_valid(self, form, *args):
        response = super().form_valid(form, *args)
        form.instance.owners.add(self.request.user)
        return response


class ScheduleUpdateView(OwnerRequiredMixin, ScheduleFormMixin, UpdateView):
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'


class ScheduleDeleteView(OwnerRequiredMixin, DeleteView):
    model = Schedule
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def get_success_url(self):
        return reverse('index')


class InvitationUpdateView(UpdateView):
    model = Invitation
    slug_field = 'token'
    slug_url_kwarg = 'token'
    fields = ['timeslot']

    def get_object(self):
        return get_object_or_404(Invitation, **self.kwargs)

    def get_success_url(self):
        return reverse('invitation', args=[self.object.schedule.uuid, self.object.token])

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        q_available = models.Q(invitation=None)
        if not self.request.user.is_authenticated:
            threshold = (
                timezone.localtime(timezone.now())
                + timedelta(hours=self.object.schedule.advance)
            )
            q_available &= (
                models.Q(date__gt=threshold.date())
                | models.Q(date=threshold.date(), time__gt=threshold.time())
            )

        timeslots = list(self.object.schedule.timeslot_set.filter(
            q_available | models.Q(invitation=self.object)
        ))

        dates = []
        times = []
        for timeslot in timeslots:
            dates.append(timeslot.date)
            times.append(timeslot.time)

        dates = sorted(set(dates))
        times = sorted(set(times))

        table = OrderedDict()
        for date in dates:
            table[date] = OrderedDict()
            for time in times:
                table[date][time] = 0
        for timeslot in timeslots:
            table[timeslot.date][timeslot.time] = timeslot.id

        context['table'] = {
            'timeslots': table,
            'times': times,
        }

        return context

    def form_valid(self, form, *args):
        timeslot = form.cleaned_data.get('timeslot')
        if timeslot and timeslot.schedule != self.object.schedule:
            messages.error(self.request, _('Invalid timeslot'))
            return self.form_invalid(form)
        response = super().form_valid(form, *args)
        messages.success(self.request, _('Your booking has been saved.'))
        return response


@method_decorator(csrf_exempt, 'dispatch')
class ScheduleApiView(View):
    def dispatch(self, request, *args, **kwargs):
        if request.headers.get('Authorization') != 'token ' + settings.API_TOKEN:
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        schedule = get_object_or_404(Schedule, **kwargs)
        return JsonResponse({
            invitation.token: invitation.datetime for invitation in schedule.invitation_set.all()
        })


@method_decorator(csrf_exempt, 'dispatch')
class InvitationApiView(View):
    def dispatch(self, request, *args, **kwargs):
        if request.headers.get('Authorization') != 'token ' + settings.API_TOKEN:
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        invitation = get_object_or_404(Invitation, **kwargs)
        return JsonResponse({'datetime': invitation.datetime})

    def put(self, request, *args, **kwargs):
        if 'schedule__uuid' in kwargs:
            schedule = get_object_or_404(Schedule, uuid=kwargs.pop('schedule__uuid'))
            kwargs['schedule_id'] = schedule.id
        invitation, _ = Invitation.objects.get_or_create(**kwargs)
        return HttpResponse(status=204)

    def delete(self, request, *args, **kwargs):
        invitation = get_object_or_404(Invitation, **kwargs)
        invitation.delete()
        return HttpResponse(status=204)
